/*
 * Copyright notice from original mutt:
 * Copyright (C) 1997 Alain Penders <Alain@Finale-Dev.com>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

#include <lib-lib/lib-lib.h>

#include <lib-sys/mutt_signal.h>
#include <lib-sys/unix.h>

#include <lib-ui/curses.h>

#include "mutt.h"

#include "mx.h"
#include "mbox.h"
#include "compress.h"

struct compress_info {
  const char *close;            /* close-hook  command */
  const char *open;             /* open-hook   command */
  const char *append;           /* append-hook command */
  off_t size;                   /* size of real folder */
};

char echo_cmd[HUGE_STRING];

/* parameters:
 * ctx - context to lock
 * excl - exclusive lock?
 * retry - should retry if unable to lock?
 */
static int mbox_lock_compressed (CONTEXT * ctx, FILE * fp, int excl, int retry)
{
  int r;

  if ((r = mx_lock_file (ctx->realpath, fileno (fp), excl, 1, retry)) == 0)
    ctx->locked = 1;
  else if (retry && !excl) {
    ctx->readonly = 1;
    return 0;
  }

  return r;
}

static void mbox_unlock_compressed (CONTEXT * ctx, FILE * fp)
{
  if (ctx->locked) {
    fflush (fp);

    mx_unlock_file (ctx->realpath, fileno (fp), 1);
    ctx->locked = 0;
  }
}

static int is_new (const char *path)
{
  return (access (path, W_OK) && errno == ENOENT);
}

static const char *find_compress_hook (int type, const char *path)
{
  const char *c = mutt_find_hook (type, path);

  return (!c || !*c) ? NULL : c;
}

int mutt_can_read_compressed (const char *path)
{
  return find_compress_hook (M_OPENHOOK, path) ? 1 : 0;
}

/* if the file is new, we really do not append, but create, and so use
 * close-hook, and not append-hook 
 */
static const char *get_append_command (const char *path, const CONTEXT * ctx)
{
  return is_new(path) ? ctx->cinfo->close : ctx->cinfo->append;
}

int mutt_can_append_compressed (const char *path)
{
  int magic;

  if (is_new (path))
    return (find_compress_hook (M_CLOSEHOOK, path) ? 1 : 0);

  magic = mx_get_magic (path);

  if (magic != 0 && magic != M_COMPRESSED)
    return 0;

  return (find_compress_hook (M_APPENDHOOK, path)
          || (find_compress_hook (M_OPENHOOK, path)
              && find_compress_hook (M_CLOSEHOOK, path))) ? 1 : 0;
}

/* open a compressed mailbox */
static compress_info *set_compress_info (CONTEXT * ctx)
{
  compress_info *ci = p_new(compress_info, 1);

  /* Now lets uncompress this thing */
  ci->append = find_compress_hook (M_APPENDHOOK, ctx->path);
  ci->open = find_compress_hook (M_OPENHOOK, ctx->path);
  ci->close = find_compress_hook (M_CLOSEHOOK, ctx->path);

  return (ctx->cinfo = ci);
}

static int get_size (const char *path)
{
  struct stat sb;

  if (stat (path, &sb) != 0)
    return 0;
  return sb.st_size;
}

static const char *
compresshook_format_str(char *dest, ssize_t destlen,
                        char op, const char *src, const char *fmt,
                        const char *ifstr __attribute__ ((unused)),
                        const char *elstr __attribute__ ((unused)),
                        anytype data,
                        format_flag flags __attribute__ ((unused)))
{
  char tmp[STRING];

  CONTEXT *ctx = data.ptr;

  switch (op) {
  case 'f':
    snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
    snprintf (dest, destlen, tmp, ctx->realpath);
    break;
  case 't':
    snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
    snprintf (dest, destlen, tmp, ctx->path);
    break;
  }
  return src;
}

/* check that the command has both %f and %t
 * 0 means OK, -1 means error
 */
int mutt_test_compress_command (const char *cmd)
{
  return (strstr (cmd, "%f") && strstr (cmd, "%t")) ? 0 : -1;
}

static char *get_compression_cmd(const char *cmd, const CONTEXT *ctx)
{
    char buf[_POSIX_PATH_MAX];
    m_strformat(buf, sizeof(buf), 0, cmd, compresshook_format_str, (void*)ctx, 0);
    return m_strdup(buf);
}

int mutt_check_mailbox_compressed (CONTEXT * ctx)
{
  if (ctx->cinfo->size != get_size (ctx->realpath)) {
    p_delete(&ctx->cinfo);
    p_delete(&ctx->realpath);
    mutt_error _("Mailbox was corrupted!");

    return -1;
  }
  return 0;
}

int mutt_open_read_compressed(CONTEXT * ctx)
{
  char *cmd;
  FILE *fp;
  int rc;
  char tmppath[_POSIX_PATH_MAX];
  int tmpfd;

  compress_info *ci = set_compress_info (ctx);

  if (!ci->open) {
    ctx->magic = 0;
    p_delete(&ctx->cinfo);
    return -1;
  }
  if (!ci->close || access (ctx->path, W_OK) != 0)
    ctx->readonly = 1;

  /* Setup the right paths */
  ctx->realpath = ctx->path;

  /* Uncompress to /tmp */
  tmpfd = m_tempfd(tmppath, sizeof(tmppath), NONULL(Tempdir), NULL);
  /* If we cannot open tempfile, that means the file already exists (!?)
   * or we are following a symlink, which is bad and insecure.
   */
  if(!tmpfd) {
      return -1;
  }
  close(tmpfd);

  ctx->path = p_dupstr(tmppath, m_strlen(tmppath));
  ctx->cinfo->size = get_size(ctx->realpath);

  if (!ctx->quiet)
    mutt_message (_("Decompressing %s..."), ctx->realpath);

  cmd = get_compression_cmd (ci->open, ctx);
  if (cmd == NULL)
    return -1;

  if ((fp = fopen (ctx->realpath, "r")) == NULL) {
    mutt_perror (ctx->realpath);
    p_delete(&cmd);
    return -1;
  }
  mutt_block_signals ();
  if (mbox_lock_compressed (ctx, fp, 0, 1) == -1) {
    m_fclose(&fp);
    mutt_unblock_signals ();
    mutt_error _("Unable to lock mailbox!");

    p_delete(&cmd);
    return -1;
  }

  endwin ();
  fflush (stdout);
  sprintf (echo_cmd, _("echo Decompressing %s..."), ctx->realpath);
  mutt_system (echo_cmd);
  rc = mutt_system (cmd);
  mbox_unlock_compressed (ctx, fp);
  mutt_unblock_signals ();
  m_fclose(&fp);

  if (rc) {
    mutt_any_key_to_continue (NULL);
    ctx->magic = 0;
    p_delete(&ctx->cinfo);
    mutt_error(_("Error executing: %s : unable to open the mailbox!\n"), cmd);
  }
  p_delete(&cmd);
  if (rc)
    return -1;

  if (mutt_check_mailbox_compressed (ctx))
    return -1;

  ctx->magic = mx_get_magic (ctx->path);

  return 0;
}

static void restore_path (CONTEXT * ctx)
{
  p_delete(&ctx->path);
  ctx->path = ctx->realpath;
}

/* remove the temporary mailbox */
static void remove_file (CONTEXT * ctx)
{
  if (ctx->magic == M_MBOX || ctx->magic == M_MMDF)
    remove (ctx->path);
}

int mutt_open_append_compressed (CONTEXT * ctx)
{
  FILE *fh;
  compress_info *ci = set_compress_info (ctx);
  char tmppath[_POSIX_PATH_MAX];

  if (!get_append_command (ctx->path, ctx)) {
    if (ci->open && ci->close)
      return mutt_open_read_compressed(ctx);

    ctx->magic = 0;
    p_delete(&ctx->cinfo);
    return -1;
  }

  /* Setup the right paths */
  ctx->realpath = ctx->path;

  /* Uncompress to /tmp */
  fh = m_tempfile(tmppath, sizeof(tmppath), NONULL(Tempdir), NULL);
  m_fclose(&fh);

  ctx->path = p_dupstr(tmppath, m_strlen(tmppath));

  ctx->magic = DefaultMagic;

  if (is_new (ctx->realpath) || 
      (ctx->magic != M_MBOX &&
       ctx->magic != M_MMDF))
      unlink(tmppath);

  /* No error checking - the parent function will catch it */

  return 0;
}

/* close a compressed mailbox */
void mutt_fast_close_compressed (CONTEXT * ctx)
{
  if (ctx->cinfo) {
    m_fclose(&ctx->fp);

    /* if the folder was removed, remove the gzipped folder too */
    if (access (ctx->path, F_OK) != 0 && !option (OPTSAVEEMPTY))
      remove (ctx->realpath);
    else
      remove_file (ctx);

    restore_path (ctx);
    p_delete(&ctx->cinfo);
  }
}

/* return 0 on success, -1 on failure */
int mutt_sync_compressed (CONTEXT * ctx)
{
  char *cmd;
  int rc = 0;
  FILE *fp;

  if (!ctx->quiet)
    mutt_message (_("Compressing %s..."), ctx->realpath);

  cmd = get_compression_cmd (ctx->cinfo->close, ctx);
  if (cmd == NULL)
    return -1;

  if ((fp = fopen (ctx->realpath, "a")) == NULL) {
    mutt_perror (ctx->realpath);
    p_delete(&cmd);
    return -1;
  }
  mutt_block_signals ();
  if (mbox_lock_compressed (ctx, fp, 1, 1) == -1) {
    m_fclose(&fp);
    mutt_unblock_signals ();
    mutt_error _("Unable to lock mailbox!");

    ctx->cinfo->size = get_size(ctx->realpath);

    p_delete(&cmd);
    return -1;
  }

  endwin ();
  fflush (stdout);
  sprintf (echo_cmd, _("echo Compressing %s..."), ctx->realpath);
  mutt_system (echo_cmd);
  if (mutt_system (cmd)) {
    mutt_any_key_to_continue (NULL);
    mutt_error (_
                ("%s: Error compressing mailbox! Original mailbox deleted, uncompressed one kept!\n"),
                ctx->path);
    rc = -1;
  }

  mbox_unlock_compressed (ctx, fp);
  mutt_unblock_signals ();
  m_fclose(&fp);

  p_delete(&cmd);

  ctx->cinfo->size = get_size(ctx->realpath);

  return rc;
}

int mutt_slow_close_compressed (CONTEXT * ctx)
{
  FILE *fp;
  const char *append;
  char *cmd;
  compress_info *ci = ctx->cinfo;

  if (!(ctx->append && ((append = get_append_command (ctx->realpath, ctx))
                        || (append = ci->close)))) {
      /* if we can not or should not append, we only have to remove the
         compressed info, because sync was already called */
    mutt_fast_close_compressed (ctx);
    return 0;
  }

  m_fclose(&ctx->fp);

  if (!ctx->quiet) {
    if (append == ci->close)
      mutt_message (_("Compressing %s..."), ctx->realpath);
    else
      mutt_message (_("Compressed-appending to %s..."), ctx->realpath);
  }

  cmd = get_compression_cmd (append, ctx);
  if (cmd == NULL)
    return -1;

  if ((fp = fopen (ctx->realpath, "a")) == NULL) {
    mutt_perror (ctx->realpath);
    p_delete(&cmd);
    return -1;
  }
  mutt_block_signals ();
  if (mbox_lock_compressed (ctx, fp, 1, 1) == -1) {
    m_fclose(&fp);
    mutt_unblock_signals ();
    mutt_error _("Unable to lock mailbox!");

    p_delete(&cmd);
    return -1;
  }

  endwin ();
  fflush (stdout);

  if (append == ci->close)
    sprintf (echo_cmd, _("echo Compressing %s..."), ctx->realpath);
  else
    sprintf (echo_cmd, _("echo Compressed-appending to %s..."),
             ctx->realpath);
  mutt_system (echo_cmd);

  if (mutt_system (cmd)) {
    mutt_any_key_to_continue (NULL);
    mutt_error (_(" %s: Error compressing mailbox!  Uncompressed one kept!\n"),
                ctx->path);
    p_delete(&cmd);
    mbox_unlock_compressed (ctx, fp);
    mutt_unblock_signals ();
    m_fclose(&fp);
    return -1;
  }

  mbox_unlock_compressed (ctx, fp);
  mutt_unblock_signals ();
  m_fclose(&fp);
  remove_file (ctx);
  restore_path (ctx);
  p_delete(&cmd);
  p_delete(&ctx->cinfo);

  return 0;
}

mx_t const compress_mx = {
    M_COMPRESSED,
    1,
    mbox_is_magic,
    mbox_check_empty,
    access,
    mutt_open_read_compressed,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
};
