/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 *
 *  Copyright © 2006 Pierre Habouzit
 */

%include {
#   include "mutt.h"
#   include "parse.h"

    static buffer_t *buffer_merge(buffer_t *A, buffer_t **B)
    {
        buffer_addbuf(A, *B);
        buffer_delete(B);
        return A;
    }

    static inline int to_control(int c) {
        return (toupper((unsigned char)c) - 'A' + 1) & 0x7f;
    }

    static buffer_t *buffer_escape(buffer_t *buf, const char *s, ssize_t len)
    {
        assert (len > 0);

        switch (*s) {
          case 'c': case 'Z':
            if (len < 2)
                break;
            buffer_addch(buf, to_control(s[1]));
            buffer_add(buf, s + 2, len - 2);
            return buf;

          case 'x':
            if (len >= 3) {
                int i = (hexval(s[1]) << 4) | hexval(s[2]);
                if (i >= 0) {
                    buffer_addch(buf, i);
                    buffer_add(buf, s + 3, len - 3);
                    return buf;
                }
            }
            break;

          case '0': case '1': case '2': case '3':
            if (len >= 3) {
                int i = (octval(s[0]) << 6) | (octval(s[1]) << 3) | octval(s[2]);
                if (i >= 0) {
                    buffer_addch(buf, i);
                    buffer_add(buf, s + 3, len - 3);
                    return buf;
                }
            }
            break;

          case 'e': buffer_addch(buf, '\e'); return buf;
          case 'f': buffer_addch(buf, '\f'); return buf;
          case 'n': buffer_addch(buf, '\n'); return buf;
          case 'r': buffer_addch(buf, '\r'); return buf;
          case 't': buffer_addch(buf, '\t'); return buf;
          case 'v': buffer_addch(buf, '\v'); return buf;

          default:
            break;
        }

        buffer_add(buf, s, len);
        return buf;
    }
}

%name rcparse
%token_prefix RCTK_
%token_type  { segment }
%start_symbol rc
%extra_argument { struct rcstate *state }

/****************************************************************************/
/* Often used                                                               */
/****************************************************************************/

escnl ::= BSLASH NL .                      { state->linenum++; }

sp ::= SPACE .
sp ::= sp SPACE .
sp ::= sp escnl SPACE .

non_nl ::= ATOM|SPACE|BQUOTE|BSLASH|SHARP|DOLLAR|DQUOTE|EQUAL|LBRACE|QUOTE|RBRACE .
non_nl_star ::= .
non_nl_star ::= non_nl_star non_nl .

to_eol_aux ::= NL .                        { state->linenum++; }
to_eol_aux ::= SHARP non_nl_star NL .      { state->linenum++; }
to_eol ::= to_eol_aux .
to_eol ::= sp to_eol_aux .

/****************************************************************************/
/* Our macro tokens                                                         */
/****************************************************************************/

/* '-quoted tokens */
%type sqtok { buffer_t* }
%destructor sqtok  { buffer_delete(&$$); }
sqtok(Z) ::= .                                 { Z = buffer_new(); }
sqtok(Z) ::= sqtok(A) escnl .                  { Z = A; }
sqtok(Z) ::= sqtok(A) BSLASH BSLASH|QUOTE(B) . { buffer_addch(Z = A, *B.s); }
sqtok(Z) ::= sqtok(A) BSLASH ATOM|SPACE|BQUOTE|SHARP|DOLLAR|DQUOTE|EQUAL|LBRACE|RBRACE(B) . {
    buffer_addch(Z = A, '\\');
    buffer_add(Z, B.s, B.len);
}
sqtok(Z) ::= sqtok(A) ATOM|SPACE|BQUOTE|SHARP|DOLLAR|DQUOTE|EQUAL|LBRACE|RBRACE(B) . {
    buffer_add(Z = A, B.s, B.len);
}

/* "-quoted tokens */
%type dqtok { buffer_t* }
%destructor dqtok  { buffer_delete(&$$); }
dqtok(Z) ::= .                             { Z = buffer_new(); }
dqtok(Z) ::= dqtok(A) escnl .              { Z = A; }
dqtok(Z) ::= dqtok(A) BSLASH non_nl(B) .   { Z = buffer_escape(A, B.s, B.len); }
dqtok(Z) ::= dqtok(A) ATOM|SPACE|SHARP|EQUAL|LBRACE|RBRACE(B)  . {
    buffer_add(Z = A, B.s, B.len);
}
dqtok(Z) ::= dqtok(A) DOLLAR ATOM(B) . {
    // FIXME
    B = B;
    Z = A;
}
dqtok(Z) ::= dqtok(A) DOLLAR LBRACE ATOM(B) RBRACE . {
    // FIXME
    B = B;
    Z = A;
}

/* unquoted tokens */
%type uqtok { buffer_t* }
%destructor uqtok { buffer_delete(&$$); }
uqtok(Z) ::= ATOM|EQUAL|LBRACE|RBRACE(A) . { buffer_add(Z = buffer_new(), A.s, A.len); }
uqtok(Z) ::= BSLASH non_nl(A) .            { buffer_add(Z = buffer_new(), A.s, A.len); }
uqtok(Z) ::= QUOTE  sqtok(A) QUOTE .       { Z = A; }
uqtok(Z) ::= DQUOTE dqtok(A) DQUOTE .      { Z = A; }

/* token accumulator */
%type token { buffer_t* }
%destructor token { buffer_delete(&$$); }
token(Z) ::= uqtok(A) .                    { Z = A; }
token(Z) ::= token(A) escnl .              { Z = A; }
token(Z) ::= token(A) uqtok(B) .           { Z = buffer_merge(A, &B); }

/****************************************************************************/
/* list of tokens                                                           */
/****************************************************************************/

%type simple_list { string_list_t * }
%destructor simple_list { string_list_wipe(&$$); }
simple_list ::= .
simple_list(L) ::= simple_list(A) sp token(B) . {
    L = string_item_new();
    L->data = buffer_unwrap(&B);
    L->next = A;
}


/****************************************************************************/
/* Entry point : the rc lines                                               */
/****************************************************************************/

rc ::= rclines .
rclines ::= .
rclines ::= rclines rcline .

rcline ::= ALTERNATIVE_ORDER simple_list(L) to_eol . {
    string_list_append(&AlternativeOrderList, string_list_rev(L));
}
rcline ::= AUTO_VIEW         simple_list(L) to_eol . {
    string_list_append(&AutoViewList, string_list_rev(L));
}
rcline ::= HDR_ORDER         simple_list(L) to_eol . {
    string_list_append(&HeaderOrderList, string_list_rev(L));
}
rcline ::= MIME_LOOKUP       simple_list(L) to_eol . {
    string_list_append(&MimeLookupList, string_list_rev(L));
}


/* vim: set indentexpr= cin: */
